/*
  EXCHANGE
() 2004-05 .. , .. 
*/

#include "exchange.h"	//  
#include "dbparams.h"	//   
#include "globals.h"	//   
#include "bracket.h"	//   
#include "dbmgr.h"
#include "mpi.h"		// Message Passing Interface
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/*
@memo     MERGE
*/
#define MERGE_CARGO(treenode, attr)	((MERGE_cargo_t *) treenode->cargo)->attr

/*
@memo     GATHER
*/
#define GATHER_CARGO(treenode, attr)	((GATHER_cargo_t *) treenode->cargo)->attr

/*
@memo     SPLIT
*/
#define SPLIT_CARGO(treenode, attr)	((SPLIT_cargo_t *) treenode->cargo)->attr

/*
@memo     SCATTER
*/
#define SCATTER_CARGO(treenode, attr)	((SCATTER_cargo_t *) treenode->cargo)->attr

/*
  MERGE
*/
typedef struct {
	int even;	//  " ?"
} MERGE_cargo_t;

/*
  SPLIT
*/
typedef struct {
	distribute_fn_t distributefn;	//     
	int breakover;					//  " ?"
} SPLIT_cargo_t;

/*
  SCATTER
*/
typedef struct {
	int port;									//   
	int dest;									//  ,    ,   SPLIT
	int breakover;								//  " ?"
	int msgtag[NODES_IN_CLUSTER_MAX];		//      
	MPI_Request midSend[NODES_IN_CLUSTER_MAX];	//     
	tupleptr_t bufSend[NODES_IN_CLUSTER_MAX];	//     
} SCATTER_cargo_t;

/*
  GATHER
*/
typedef struct {
	int port;									//   
	int gotEOF[NODES_IN_CLUSTER_MAX];			//   " EOF?"
	int cntEOF;									//   EOF
	MPI_Request midRecv[NODES_IN_CLUSTER_MAX];	//     
	int msgtag[NODES_IN_CLUSTER_MAX];			//      
	tupleptr_t bufRecv[NODES_IN_CLUSTER_MAX];	//     
} GATHER_cargo_t;


//   -  EXCHANGE

/*
  MERGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeMERGE(int ary, int fragmentattr, fragment_fn_t fragmentfn);

/*
  GATHER
@param port    EXCHANGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeGATHER(int port, int ary, int fragmentattr, fragment_fn_t fragmentfn);

/*
  SPLIT
@param distributefn	   
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeSPLIT(distribute_fn_t distributefn, int ary, int fragmentattr, 
							  fragment_fn_t fragmentfn);

/*
  SCATTER
@param port    EXCHANGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeSCATTER(int port, int ary, int fragmentattr, fragment_fn_t fragmentfn);

// ,    -  EXCHANGE

/*
  MERGE
@param node     
@return       MERGE  , 
			   --   ;
		EOF -   ; 
		NULL -     NEXT.
*/
static tupleptr_t NextMERGE(treenode_t * node);

/*
  MERGE
@param node     
@return SUCCESS       
*/
static void ResetMERGE(treenode_t * node);

/*
  GATHER
@param node     
@return       GATHER  , 
			   --   ;
		EOF -   ; 
		NULL -     NEXT.
*/
static tupleptr_t NextGATHER(treenode_t * node);

/*
  GATHER
@param node     
*/
static void ResetGATHER(treenode_t * node);

/*
  SPLIT
@param node     
@return       SPLIT  , 
			   --   ;
		EOF -   ; 
		NULL -     NEXT.
*/
static tupleptr_t NextSPLIT(treenode_t * node);

/*
  SPLIT
@param node     
@return SUCCESS       
*/
static void ResetSPLIT(treenode_t * node);

/*
  SCATTER
@param node     
@return       SCATTER  , 
			   --   ;
		EOF -   ; 
		NULL -     NEXT.
*/
static tupleptr_t NextSCATTER(treenode_t * node);

/*
  SCATTER
@param node     
*/
static void ResetSCATTER(treenode_t * node);

//  

/*
    .
 ,         
 -       -   , 
    :
 -     -- MessageTag(__, , )
 -     -- MessageTag(, __, ).
@param source  - 
@param dest  - 
@param portEXCHANGE    EXCHANGE,      
@return  
*/
static int MessageTag(int source, int dest, int portEXCHANGE);

/*
    .
@param dest  - 
@param portEXCHANGE    EXCHANGE,      
@return  
*/
#define SEND_TAG(dest, portEXCHANGE)	MessageTag(WHOAMI, dest, portEXCHANGE)

/*
    .
@param source  - 
@param portEXCHANGE    EXCHANGE,      
@return  
*/
#define RECV_TAG(source, portEXCHANGE)	MessageTag(source, WHOAMI, portEXCHANGE)

/*
    
@param node    
@param tuple   
*/
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

// 

/*
  EXCHANGE
@param port    EXCHANGE
@param distributefn	   
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
treenode_t * MakeEXCHANGE(int port, distribute_fn_t distributefn, int ary, int fragmentattr, 
							 fragment_fn_t fragmentfn)
{
	treenode_t *Merge, *Gather, *Split, *Scatter;

	Merge=MakeMERGE(ary, fragmentattr, fragmentfn);
	assert(Merge!=NULL);
	Gather=MakeGATHER(port, ary, fragmentattr, fragmentfn);
	assert(Gather!=NULL);
	Split=MakeSPLIT(distributefn, ary, fragmentattr, fragmentfn);
	assert(Split!=NULL);
	Scatter=MakeSCATTER(port, ary, fragmentattr, fragmentfn);
	assert(Scatter!=NULL);
	Merge->left=Gather;
	Merge->right=Split;
	Split->right=Scatter;
	return Merge;
}

/*
  MERGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeMERGE(int ary, int fragmentattr, fragment_fn_t fragmentfn)
{
	treenode_t *Merge;

	assert(ary>0);
	assert(fragmentattr>=0 && fragmentattr<ATTRIBUTES_IN_RELATION);
	assert(fragmentfn!=NULL);

	Merge=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Merge!=NULL);
	Merge->rel_operation=MERGE;
	Merge->fragmentattr=fragmentattr;
	Merge->fragmentfn=fragmentfn;
	Merge->ary=ary;
	Merge->nextfn=(next_fn_t) NextMERGE;
	Merge->resetfn=(reset_fn_t) ResetMERGE;
	Merge->isBOF=Merge->isEOF=FALSE;
	Merge->left=Merge->right=NULL;
	Merge->cargo=(MERGE_cargo_t *) calloc(1, sizeof(MERGE_cargo_t));
	assert(Merge->cargo!=NULL);
	MERGE_CARGO(Merge, even)=FALSE;
	return Merge;
}

/*
  GATHER
@param port    EXCHANGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeGATHER(int port, int ary, int fragmentattr, fragment_fn_t fragmentfn)
{
	treenode_t *Gather;
	int i;

	assert(port>=0);
	assert(ary>0);
	assert(fragmentattr>=0 && fragmentattr<ATTRIBUTES_IN_RELATION);
	assert(fragmentfn!=NULL);

	Gather=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Gather!=NULL);
	Gather->rel_operation=GATHER;
	Gather->fragmentattr=fragmentattr;
	Gather->fragmentfn=fragmentfn;
	Gather->ary=ary;
	Gather->nextfn=(next_fn_t) NextGATHER;
	Gather->resetfn=(reset_fn_t) ResetGATHER;
	Gather->isBOF=Gather->isEOF=FALSE;
	Gather->left=Gather->right=NULL;
	Gather->cargo=(GATHER_cargo_t *) calloc(1, sizeof(GATHER_cargo_t));
	assert(Gather->cargo!=NULL);
	GATHER_CARGO(Gather, port)=port;
	GATHER_CARGO(Gather, cntEOF)=0;
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==WHOAMI) {
			GATHER_CARGO(Gather, bufRecv[i])=NULL;
			GATHER_CARGO(Gather, midRecv[i])=MPI_REQUEST_NULL;
			GATHER_CARGO(Gather, gotEOF[i])=FALSE;
			GATHER_CARGO(Gather, msgtag[i])=NIL;
			continue;
		}
		GATHER_CARGO(Gather, gotEOF[i])=FALSE;
		GATHER_CARGO(Gather, msgtag[i])=RECV_TAG(i, port);
//      !
		assert(GATHER_CARGO(Gather, msgtag[i])==MessageTag(i, WHOAMI, port)); 
		GATHER_CARGO(Gather, bufRecv[i])=(tupleptr_t) calloc(Gather->ary, sizeof(domain_t));
		assert(GATHER_CARGO(Gather, bufRecv[i])!=NULL);
	}
	return Gather;
}

/*
  SPLIT
@param distributefn	   
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeSPLIT(distribute_fn_t distributefn, int ary, int fragmentattr, fragment_fn_t fragmentfn)
{
	treenode_t *Split;

	assert(distributefn!=NULL);
	assert(ary>0);
	assert(fragmentattr>=0 && fragmentattr<ATTRIBUTES_IN_RELATION);
	assert(fragmentfn!=NULL);

	Split=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Split!=NULL);
	Split->rel_operation=SPLIT;
	Split->fragmentattr=fragmentattr;
	Split->fragmentfn=fragmentfn;
	Split->ary=ary;
	Split->nextfn=(next_fn_t) NextSPLIT;
	Split->resetfn=(reset_fn_t) ResetSPLIT;
	Split->isBOF=Split->isEOF=FALSE;
	Split->left=Split->right=NULL;
	Split->cargo=(SPLIT_cargo_t *) calloc(1, sizeof(SPLIT_cargo_t));
	assert(Split->cargo!=NULL);
	SPLIT_CARGO(Split, breakover)=FALSE;
	SPLIT_CARGO(Split, distributefn)=distributefn;
	return Split;
}

/*
  SCATTER
@param port    EXCHANGE
@param ary   
@param fragmentattr  
@param fragmentfn    
@return         NULL   
*/
static treenode_t * MakeSCATTER(int port, int ary, int fragmentattr, fragment_fn_t fragmentfn)
{
	treenode_t *Scatter;
	int i;

	assert(port>=0);
	assert(ary>0);
	assert(fragmentattr>=0 && fragmentattr<ATTRIBUTES_IN_RELATION);
	assert(fragmentfn!=NULL);

	Scatter=(treenode_t *) calloc(1, sizeof(treenode_t));
	assert(Scatter!=NULL);
	Scatter->rel_operation=SCATTER;
	Scatter->fragmentattr=fragmentattr;
	Scatter->fragmentfn=fragmentfn;
	Scatter->ary=ary;
	Scatter->nextfn=(next_fn_t) NextSCATTER;
	Scatter->resetfn=(reset_fn_t) ResetSCATTER;
	Scatter->isBOF=Scatter->isEOF=FALSE;
	Scatter->left=Scatter->right=NULL;
	Scatter->cargo=(SCATTER_cargo_t *) calloc(1, sizeof(SCATTER_cargo_t));
	assert(Scatter->cargo!=NULL);
	SCATTER_CARGO(Scatter, port)=port;
	SCATTER_CARGO(Scatter, breakover)=FALSE;
	SCATTER_CARGO(Scatter, dest)=NIL; 
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==WHOAMI) {
			SCATTER_CARGO(Scatter, bufSend[i])=NULL;
			SCATTER_CARGO(Scatter, midSend[i])=MPI_REQUEST_NULL;
			SCATTER_CARGO(Scatter, msgtag[i])=NIL;
			continue;
		}
		SCATTER_CARGO(Scatter, msgtag[i])=SEND_TAG(i, port);
//      !
		assert(SCATTER_CARGO(Scatter, msgtag[i])==MessageTag(WHOAMI, i, port)); 
		SCATTER_CARGO(Scatter, bufSend[i])=(tupleptr_t) calloc(Scatter->ary, sizeof(domain_t));
		assert(SCATTER_CARGO(Scatter, bufSend[i])!=NULL);
		SCATTER_CARGO(Scatter, midSend[i])=MPI_REQUEST_NULL;
	}
	return Scatter;
}

/*
     MERGE (  EXCHANGE). 
      " "   
 .
@memo   MERGE
@return      , EOF   ,
	NULL --    .
*/
static tupleptr_t NextMERGE(treenode_t * node)
{
	tupleptr_t tl=NULL, tr=NULL;

	while (TRUE) {
		MERGE_CARGO(node, even)=!MERGE_CARGO(node, even);
		switch (MERGE_CARGO(node, even)) {
		case FALSE: //     SPLIT ( ),   GATHER ( )
			tr=br_Next(node->right);//    SPLIT
			if (!IsEOF(tr) && tr!=NULL) {
				RESCPY(node, tr);
				return node->result;
			} 
			tl=br_Next(node->left);//    GATHER
			if (!IsEOF(tl) && tl!=NULL) {
				RESCPY(node, tl);
				return node->result;
			}
			break;
		case TRUE: //     GATHER ( ),   SPLIT ( )
			tl=br_Next(node->left); //    GATHER
			if (!IsEOF(tl) && tl!=NULL) {
				RESCPY(node, tl);
				return node->result;
			}
			tr=br_Next(node->right);//    SPLIT
			if (!IsEOF(tr) && tr!=NULL) {
				RESCPY(node, tr);
				return node->result;
			}
			break;
		default:
			assert(0);
		}
		if (IsEOF(tl) && IsEOF(tr))
			break;
	} 
	RESCPY(node, EOFtuple);
	return node->result;
}

/*
  MERGE
@param node     
*/
static void ResetMERGE(treenode_t * node)
{
	assert(node!=NULL);
	MERGE_CARGO(node, even)=FALSE;
	return;
}

/*
     SPLIT (  EXCHANGE). 
      " "   
 .
@memo   SPLIT
@return      , EOF   ,
	NULL --    .
*/
static tupleptr_t NextSPLIT(treenode_t * node)
{
	int dest;
	tupleptr_t result;

	assert(node!=NULL);

	if (SPLIT_CARGO(node, breakover)) { //      SCATTER  
		if (br_Next(node->right)!=NULL) //   
			SPLIT_CARGO(node, breakover)=FALSE;
		return NULL;
	}
	
	if (IsEOF(node->result))  //  , ..  EOF   
		return node->result;
	
	//    
	result=br_Next(node->left);
	assert(result!=NULL);
	RESCPY(node, result);
	if (IsEOF(result)) { //   EOF
		RESCPY(node->right, result); //  EOF  .  SCATTER
		if (br_Next(node->right)==NULL) //  EOF  
			SPLIT_CARGO(node, breakover)=TRUE;
		return NULL;	
	}

	dest=SPLIT_CARGO(node, distributefn)(result, node->fragmentattr); //  
	assert(dest>=0 && dest<NODES_IN_CLUSTER);

	if (dest==WHOAMI)  //     --   
		return node->result;
	
	//  -- 
	SCATTER_CARGO(node->right, dest)=dest; //  SCATTER,     
	RESCPY(node->right, result);	 //    .  SCATTER
	if (br_Next(node->right)==NULL) 
		SPLIT_CARGO(node, breakover)=TRUE;
	return NULL;
}

/*
  SPLIT
@param node     
@see reset_fn_t NextSPLIT
*/
static void ResetSPLIT(treenode_t * node)
{
	assert(node!=NULL);
	SPLIT_CARGO(node, breakover)=FALSE;
	return;
}

/*
     GATHER (  EXCHANGE). 
      " "   
 .
@memo   GATHER
@return       GATHER  , 
			   --   ;
		EOF -   ; 
		NULL -     NEXT.
*/
static tupleptr_t NextGATHER(treenode_t * node)
{
	int i, res, flag;
	MPI_Status status;

	assert(node!=NULL);
	if (GATHER_CARGO(node, cntEOF)==NODES_IN_CLUSTER-1) //  EOF   
		return EOFtuple;

	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==WHOAMI || GATHER_CARGO(node, gotEOF[i]))
			continue;
		res=MPI_Test(&GATHER_CARGO(node, midRecv[i]), &flag, &status);
		assert(res==MPI_SUCCESS);
		if (flag) {
			if (IsEOF(GATHER_CARGO(node, bufRecv[i]))) { //  EOF 
				GATHER_CARGO(node, gotEOF[i])=TRUE;
				GATHER_CARGO(node, cntEOF)++;
				return NULL;
			};
			//   EOF 
			RESCPY(node, GATHER_CARGO(node, bufRecv[i]));
			/* 
				    
				int MPI_Irecv(void *buf, int count, MPI_Datatype datatype, 
						int source, int msgtag, MPI_comm comm, MPI_Request *request) 
			*/
			res=MPI_Irecv(GATHER_CARGO(node, bufRecv[i]), node->ary*sizeof(domain_t), MPI_BYTE, 
							i, GATHER_CARGO(node, msgtag[i]), MPI_COMM_WORLD, 
							&GATHER_CARGO(node, midRecv[i]));
			assert(res==MPI_SUCCESS);
			return node->result;
		}
	}
	return NULL;
}

/*
  GATHER
@param node     
*/
static void ResetGATHER(treenode_t * node)
{
	int i, res;

	assert(node!=NULL);
	GATHER_CARGO(node, cntEOF)=0; //   EOF
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==WHOAMI) 
			continue;		
		GATHER_CARGO(node, gotEOF[i])=FALSE;
		/*
		int MPI_Irecv(void *buf, int count, MPI_Datatype datatype, 
					int source, int msgtag, MPI_comm comm, MPI_Request *request) 
		*/
		res=MPI_Irecv(GATHER_CARGO(node, bufRecv[i]), node->ary*sizeof(domain_t), MPI_BYTE, i, 
						GATHER_CARGO(node, msgtag[i]), MPI_COMM_WORLD, &GATHER_CARGO(node, midRecv[i]));
		assert(res==MPI_SUCCESS);
	}
	return;
}

/*
     SCATTER (  EXCHANGE). 
      " "   
 .
@memo   SCATTER
@return      , EOF   ,
	NULL --    .
*/
/*
static tupleptr_t NextSCATTER(treenode_t * node)	//    
{
	int i, 
		res, //  MPI- 
		flag=TRUE;//     MPI

	if (IsEOF(node->result)) { //     SCATTER -- EOF,     EOF
		SCATTER_CARGO(node, dest)=WHOAMI; //    assert(res==MPI_SUCCESS); ( )
		for (i=0; i<NODES_IN_CLUSTER; i++) {
			if (i==WHOAMI) 
				continue;
			memcpy(SCATTER_CARGO(node, bufSend[i]), node->result, node->ary*sizeof(domain_t));
			res=MPI_Send(SCATTER_CARGO(node, bufSend[i]), node->ary*sizeof(domain_t), MPI_BYTE, i, 
				SCATTER_CARGO(node, msgtag[i]), MPI_COMM_WORLD);
			assert(res==MPI_SUCCESS);
		}
	} else { //     SCATTER --  EOF,   -  
		assert(SCATTER_CARGO(node, dest)!=NIL);
		memcpy(SCATTER_CARGO(node, bufSend[SCATTER_CARGO(node, dest)]), node->result, 
				node->ary*sizeof(domain_t));
		res=MPI_Send(SCATTER_CARGO(node, bufSend[SCATTER_CARGO(node, dest)]), node->ary*sizeof(domain_t), 
				MPI_BYTE, SCATTER_CARGO(node, dest), 
				SCATTER_CARGO(node, msgtag[SCATTER_CARGO(node, dest)]), MPI_COMM_WORLD);
		assert(res==MPI_SUCCESS);
	}
	return node->result;
}
*/
static tupleptr_t NextSCATTER(treenode_t * node)	//    
{
	int i, 
		res, //  MPI- 
		flag=TRUE;//     MPI
	MPI_Status status;

	assert(node!=NULL);
	if (SCATTER_CARGO(node, breakover)) {
		if (IsEOF(node->result)) { //    EOF  
			for (i=0; i<NODES_IN_CLUSTER; i++) {
				if (i==WHOAMI) 
					continue;
				res=MPI_Test(&SCATTER_CARGO(node, midSend[i]), &flag, &status);
				assert(SCATTER_CARGO(node, dest)!=NIL && res==MPI_SUCCESS);
				if (!flag)  //  EOF  
					return NULL; 
			}
			//  EOF 
			SCATTER_CARGO(node, breakover)=FALSE;
			return node->result; 
		} else {//    
			res=MPI_Test(&SCATTER_CARGO(node, midSend[SCATTER_CARGO(node, dest)]), &flag, &status);
			assert(SCATTER_CARGO(node, dest)!=NIL && res==MPI_SUCCESS);
			if (!flag)  //     
				return NULL; 
			//   
			SCATTER_CARGO(node, breakover)=FALSE;
			return node->result; 
		}
	}
	
	SCATTER_CARGO(node, breakover)=TRUE;

	if (IsEOF(node->result)) { //     SCATTER -- EOF,     EOF
		SCATTER_CARGO(node, dest)=WHOAMI; //    assert(res==MPI_SUCCESS); ( )
		for (i=0; i<NODES_IN_CLUSTER; i++) {
			if (i==WHOAMI) 
				continue;
			memcpy(SCATTER_CARGO(node, bufSend[i]), node->result, node->ary*sizeof(domain_t));
			res=MPI_Isend(SCATTER_CARGO(node, bufSend[i]), node->ary*sizeof(domain_t), MPI_BYTE, i, 
				SCATTER_CARGO(node, msgtag[i]),	MPI_COMM_WORLD, &SCATTER_CARGO(node, midSend[i]));
			assert(res==MPI_SUCCESS);
		}
	} else { //     SCATTER --  EOF,   -  
		assert(SCATTER_CARGO(node,dest)!=NIL);
		memcpy(SCATTER_CARGO(node, bufSend[SCATTER_CARGO(node,dest)]), node->result, 
			node->ary*sizeof(domain_t));
		res=MPI_Isend(SCATTER_CARGO(node, bufSend[SCATTER_CARGO(node, dest)]), node->ary*sizeof(domain_t), 
				MPI_BYTE, SCATTER_CARGO(node, dest), SCATTER_CARGO(node, msgtag[SCATTER_CARGO(node, dest)]), 
				MPI_COMM_WORLD, &SCATTER_CARGO(node, midSend[SCATTER_CARGO(node, dest)]));
		assert(res==MPI_SUCCESS);
	}
	return NULL;
}

/*
  SCATTER
@param node     
*/
static void ResetSCATTER(treenode_t * node)
{
	int i;

	assert(node!=NULL);
	SCATTER_CARGO(node, dest)=NIL;
	SCATTER_CARGO(node, breakover)=FALSE;
	for (i=0; i<NODES_IN_CLUSTER; i++) {
		if (i==WHOAMI) 
			continue;
		SCATTER_CARGO(node, midSend[i])=MPI_REQUEST_NULL;
	}
	return;
}

/*
    .
 ,         
 -       -   , 
    :
       -- MessageTag(__, , )
       -- MessageTag(, __, ).
@param source  - 
@param dest  - 
@param portEXCHANGE    EXCHANGE,      
@return  
*/
static int MessageTag(int source, int dest, int portEXCHANGE)
{
	int tag;
	
	assert(source>=0 && source<NODES_IN_CLUSTER);
	assert(dest>=0 && dest<NODES_IN_CLUSTER);
	assert(source!=dest);
	assert(portEXCHANGE>=0);
	tag=portEXCHANGE*10000 + source*100 + dest;
	assert(tag>=0 && tag<32767);
	return tag;
}
