/*
  JOIN ( )  , 
       .
.. 
@version 1.0
*/

#include "bracket.h"	//  
#include "omegacfg.h"
#include "jgrace.h"
#include "dbmgr.h"
#include <string.h>
#include <stdlib.h> 
#include <stdio.h>
#include "assert.h"
#include "math.h"

/*
 *  :
 *
 *   ,     - ,      
 *   .
 *        .
 *
 *   :
 *
 * 1)    -      .
 * 2)  ,     .
 * 3)          .
 * 4)        .
 * 5)       .
 * 6)          ,    . 2.
 * 7)    .
 * 8)   ""    .
 * 9)       ,        
 *      -.
 * 10)     .
 * 11)     ,    . 8.
 *
 * 
 *
 *       :
 * 1)    .
 * 2)      .
 * 3)       (  )
*/
/* INTERFACE */

/* Definitions: */

/*    src  dest */
#define TUPLECPY(dest, src, ary) memcpy(dest, src, ary*sizeof(domain_t))

/*    tuple  node->result */
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

/*  ary   src  dest(     ary   !). */
#define BYTECPY(dest, src, ary) memcpy(dest, src, ary)

/* Types: */

/*
 * .       .
 * filesize -  ,    -.
 * Probefilesize -  ,    -.
 * max -    .
 * min -    .
 * samplingsize -     .
 * sampling[] -     (   ,    ).
*/
typedef struct {
	long filesize;
	long Probefilesize;
	long max,min, samplingsize;
	long sampling[JOIN_GRACE_SAMPLING_BUFLEN_MAX];
} StatInfo_t;

/*
 *   -.
 * tuple -   .
 * next -         .
*/
struct _buf{
	tuple_t tuple;
	struct _buf *next;
};
typedef struct _buf buf_t;

/*
 *     .
 * Filename -   ,   ,   
 *   -.
 * ProbeFilename -   ,   ,   
 *    -.
 * statinfo -     .
 * ReverseSign -  ,    ,     
 * (  
 *     ,        
 *    .)
 * handle -   ,   ,    
 *  -.
 * Probehandle -   ,   ,   
 *    -.
 * top, bottom -          , 
 *    .
 * 
*/
typedef struct {
	char Filename[FILENAME_LEN_MAX + PATHNAME_LEN_MAX];
	char ProbeFilename[FILENAME_LEN_MAX + PATHNAME_LEN_MAX];
	StatInfo_t statinfo;
	long top, bottom;
	int ReverseSign;
	FILE *handle;
	FILE *Probehandle;
} ListFiles_t;
/*
 *  -.
 * buffer -   -.
 * size -  -.
 * current -  -,        .
*/
typedef struct {
	buf_t buffer[JOIN_GRACE_BUFLEN_MAX];	
	long size;
	buf_t *current;
} TupleBuffer_t;

/*
 *   GRACE - .
 * table -   ,  -.
 * listfiles -     .
 * ResetSign -   nextJOIN_Grace      .
 * WhoAmI -  .
 * Sizelistfiles -   .
 * currentfile -   .
 * Probetuple -    "" .
*/
typedef struct {
	TupleBuffer_t table;	
	ListFiles_t listfiles[JOIN_GRACE_TEMPFILES_MAX];
	int ResetSign;
	int WhoAmI;
	long Sizelistfiles;
	long currentfile;
	tuple_t Probetuple;
} JOIN_GRACE_cargo_t;
#define JOIN_GRACE_CARGO(treenode, attr)	((JOIN_GRACE_cargo_t *) JOIN_CARGO(treenode, supercargo))->attr

/* Variables: */
/*    .      . */
long Counter = 0;

/* Function prototypes: */
/**
       
   JOIN_GRACE
(  ,       
    ). 
@memo   JOIN_GRACE
@param node   
@see JOIN_GRACE _treenode_t JOIN_GRACE_BUFLEN_MAX
*/
void AssignJOIN_Grace(treenode_t * node);

/*   */
static void ResetJOIN_Grace(treenode_t * node);

/*    . */
static tupleptr_t NextJOIN_Grace(treenode_t * node);

/*  -. */
void maketable(TupleBuffer_t *table, ListFiles_t *list, long current, int ary, int attr);

/* IMPLEMENTATION */

void AssignJOIN_Grace(treenode_t * node)
{
	assert(node!=NULL);
	assert(node->rel_operation==JOIN);
	
	node->phys_operation=JOIN_GRACE;
	node->nextfn=(next_fn_t) NextJOIN_Grace;
	node->resetfn=(reset_fn_t) ResetJOIN_Grace;
	JOIN_CARGO(node, supercargo)=(JOIN_GRACE_cargo_t *) calloc(1, sizeof(JOIN_GRACE_cargo_t));
	assert(JOIN_CARGO(node, supercargo)!=NULL);

	JOIN_GRACE_CARGO(node, ResetSign) = TRUE;
	JOIN_GRACE_CARGO(node, WhoAmI) = WHOAMI;
	JOIN_GRACE_CARGO(node, Sizelistfiles) = 0;
	
	return;
}
/*
 *  ,       
 *      list  size,      
 *    , 
 *   number.
*/
long getpos(ListFiles_t *list, long number, long size)
{
	long pos;

	pos = 0;
	/* !       !*/
	while (!((number >= list[pos].bottom) && (number <= list[pos].top))) {
		pos++;
		if (pos == size)
			return NIL;
	}

	return pos;
}

/*
 *    .
 *         .
*/
long jg_partition(long *buffer, long p, long r)

{
	long x;
	long temp;// = (tuple_t) calloc(ary, sizeof(domain_t));
	long i = p - 1;
	long j = r + 1;
	long k = rand()%(r-p) + p;
	
	temp      = buffer[k];
	buffer[k] = buffer[p];
	buffer[p] = temp;
	x         = buffer[p];

	while (1==1) {
		do
			j--;
		while ((buffer[j] >= x) && (j>p));

		do
			i++;
		while ((buffer[i] < x) && (i<r));

			if (i<j) {
				temp      = buffer[i];
				buffer[i] = buffer[j];
				buffer[j] = temp;
			}
			else
				return j;
	}

}

/*
 *       buffer    [p..r].
*/
void jg_quicksort(long *buffer, long p, long r)
{
	long q;
	if (p < r) {
		q = jg_partition (buffer, p, r);
		jg_quicksort (buffer, p, q);
		jg_quicksort (buffer, q+1, r);
	}
	return;
}

/* 
 *      (   ).
 *       Filename.     
 *  ,      WhoAmi
 *   Filename     .
*/
char * jg_CreateFilename(int WhoAmi, char *Filename)
{
	char sn[12] = "";

	Counter++;
	BYTECPY(Filename, PATH_TO_RELATIONS, strlen(PATH_TO_RELATIONS)+1);
	sprintf(sn, "%s", "N");
	strcat(Filename, sn);
	sprintf(sn, "%d", WhoAmi);
	strcat(Filename, sn);
	sprintf(sn, "%s", "C");
	strcat(Filename, sn);
	sprintf(sn, "%ld", Counter);
	strcat(Filename,sn);
	strcat(Filename, ".tmp");
	return Filename;
} /* jg_CreateFilename() */

/*
 *   tuple,  ary     handle.
*/
int jg_gettuple(FILE *handle, tupleptr_t tuple, int ary)
{
	int res, i;

	for (i=0; i<ary; i++) {
		res = fscanf(handle, "%ld\t", &tuple[i]);
	}
	res = fscanf(handle, "\n");

	return res;
}
/*
 *   tuple,  ary     handle.
*/
int jg_puttuple(FILE *handle, tupleptr_t tuple, int ary)
{
	int res, i;

	for (i=0; i<ary; i++) {
		res = fprintf(handle, "%ld\t", tuple[i]);
	}
	res = fprintf(handle, "\n");

	return res;
}

/*
 *   .           
 *   .
*/
void jg_prepairing(treenode_t *node, JOIN_GRACE_cargo_t *Descriptor, int attr)
{
	FILE *handle;
	char filename[FILENAME_LEN_MAX + PATHNAME_LEN_MAX];
	int res;

	br_Next(node);
	
	if (node->isEOF) //    ...
		return;
	// 
	jg_CreateFilename(Descriptor->WhoAmI, filename);
	handle = fopen(filename, "wt");
	assert(handle != NULL);
	Descriptor->listfiles[0].statinfo.filesize = 0;
	Descriptor->listfiles[0].statinfo.max = node->result[attr];
	Descriptor->listfiles[0].statinfo.min = node->result[attr];
	Descriptor->listfiles[0].statinfo.samplingsize = 0;
	///////////////
	//   (        ).
	while (!node->isEOF)
	{
		res = jg_puttuple(handle, node->result, node->ary);
		assert(res != EOF);
		//  . ...
		Descriptor->listfiles[0].statinfo.filesize++;
		if (Descriptor->listfiles[0].statinfo.min > node->result[attr])
			Descriptor->listfiles[0].statinfo.min = node->result[attr];
		if (Descriptor->listfiles[0].statinfo.max < node->result[attr])
			Descriptor->listfiles[0].statinfo.max = node->result[attr];
		if (((rand()%3) == 1) && (Descriptor->listfiles[0].statinfo.samplingsize < JOIN_GRACE_SAMPLING_BUFLEN_MAX))
		{
			Descriptor->listfiles[0].statinfo.sampling[Descriptor->listfiles[0].statinfo.samplingsize] = node->result[attr];
			Descriptor->listfiles[0].statinfo.samplingsize++;
		}
		/////////////////////////
		br_Next(node);
	}
	res = fclose(handle);
	assert(res != EOF);
	BYTECPY(Descriptor->listfiles[0].Filename, filename, strlen(filename)+1);
	Descriptor->Sizelistfiles = 1;
	Descriptor->listfiles[0].top = Descriptor->listfiles[0].statinfo.max+1;
	Descriptor->listfiles[0].bottom = Descriptor->listfiles[0].statinfo.min;
	//     (      
	//   )
	jg_quicksort(Descriptor->listfiles[0].statinfo.sampling, 0, Descriptor->listfiles[0].statinfo.samplingsize-1);

	return;
}

#define stat(attr) Descriptor->listfiles[current].statinfo.attr

/*
 *             
 * Descriptor->listfiles[current].
*/
int createrangefiles(JOIN_GRACE_cargo_t *Descriptor, ListFiles_t *list, long current)
{
	long amount, step, count, k, i;
	//   ,      .
	amount = stat(filesize) /  (long)floor(((float)(Descriptor->table.size)/100) * 70) + 1;
	assert(amount < JOIN_GRACE_TEMPSUBFILES_MAX);
	//  .       . step  
	//         
	//  .
	step   = (stat(samplingsize) / amount) + 1;

	// 
	for (i = 0; i<amount; i++) {
		list[i].statinfo.filesize = 0;
		list[i].statinfo.Probefilesize = 0;
		list[i].statinfo.min = NIL;
		list[i].statinfo.max = NIL;
		list[i].ReverseSign = TRUE;
		list[i].statinfo.samplingsize = 0;
		jg_CreateFilename(Descriptor->WhoAmI, list[i].Filename);
	}
	i = 0;
	k = 0;
	count = 0;
	////////////////
	//    .
	while (count < stat(samplingsize-1))	{
		if (k >= step) {
			while ( (stat(sampling[count+1]) == stat(sampling[count])) && (count+2 != stat(samplingsize)) )
				count++;
			if (i>0)
				list[i].bottom = list[i-1].top+1;
			else
				list[0].bottom = stat(min);
			list[i].top = (stat(sampling[count+1]) + stat(sampling[count]))/2;

			k = 0;
			i++;
		}
		k++;
		count++;
	}
	//       ,    ,
	//   .
	for (; i < amount; i++)	{
		if (i>0)
			list[i].bottom = list[i-1].top+1;
		else
			list[0].bottom = stat(min);
		list[i].top = list[i].bottom + (stat(max) - list[i].bottom)/(amount-i);
	}
	list[amount-1].top = stat(max);

	return amount;
} /* createranges() */

#undef stat

/*
 *    ary     handle     list, 
 *    attr.
*/
void distrib_tuples(ListFiles_t *list, FILE *handle, int ary, int attr, long amount)
{
	tuple_t tuple;
	int res, i;

	//    .
	for (i = 0; i < amount; i++) {
		list[i].handle = fopen(list[i].Filename, "wt");
		assert(list[i].handle != NULL);
	}
	//  
	while (!feof(handle))	{

		res = jg_gettuple(handle, tuple, ary);
			
		i = getpos(list, tuple[attr], amount);
		assert(i != NIL);

		res = jg_puttuple(list[i].handle, tuple, ary);
		assert(res != EOF);

		list[i].statinfo.filesize++;

		if (list[i].statinfo.min == NIL) {
			list[i].statinfo.min = tuple[attr];
			list[i].statinfo.max = tuple[attr];
		}
		if (list[i].statinfo.min > tuple[attr])
			list[i].statinfo.min = tuple[attr];
		if (list[i].statinfo.max < tuple[attr])
			list[i].statinfo.max = tuple[attr];

		if (((rand()%3) == 1) && (list[i].statinfo.samplingsize < JOIN_GRACE_SAMPLING_BUFLEN_MAX)) {
			list[i].statinfo.sampling[list[i].statinfo.samplingsize] = tuple[attr];
			list[i].statinfo.samplingsize++;
		}
	} // while
	//  
	//     .
	for (i = 0; i< amount; i++) {
		jg_quicksort(list[i].statinfo.sampling, 0, list[i].statinfo.samplingsize-1);
	}
	//   .
	for (i = 0; i < amount; i++) {
		res = fclose(list[i].handle);
		assert(res != EOF);
	}

	return;
}

/*       size,   -. */
long getsizetable(long size)
{
	long m;
	
	m = 1;
	while (m <= size)
		m = m<<1;
	m = m>>1;
	assert (m != 0);

	return m;
}

/*
 *       .
 *   ,         .
 *       .
*/
void listoptimizer(ListFiles_t *list, long *size)
{
	long i = 0, j = 0;
	char filename[FILENAME_LEN_MAX + PATHNAME_LEN_MAX];
	long temp;
	
	while (i < *size) {
		if (list[i].statinfo.filesize > list[i].statinfo.Probefilesize) {
			BYTECPY(filename, list[i].Filename, strlen(list[i].Filename)+1);
			BYTECPY(list[i].Filename, list[i].ProbeFilename, strlen(list[i].ProbeFilename)+1);
			BYTECPY(list[i].ProbeFilename, filename, strlen(filename)+1);
			temp = list[i].statinfo.filesize;
			list[i].statinfo.filesize = list[i].statinfo.Probefilesize;
			list[i].statinfo.Probefilesize = temp;
			list[i].ReverseSign = FALSE;
		}

		if ((list[i].statinfo.filesize == 0) || (list[i].statinfo.Probefilesize == 0)) {
			j = i;
			while (j < *size-1) {
				BYTECPY(&list[j], &list[j+1], sizeof(list[j+1]));
				j++;
			}
			*size = *size-1;
			i = 0;
		}
		i++;
	}
	return;
}

/*
 *      ,      .
 *       .     
 *    
 *    .
 *   . 
*/
void jg_partitioning(treenode_t *node, JOIN_GRACE_cargo_t *Descriptor, int attr)
{
	FILE *handle;
	char filename[FILENAME_LEN_MAX + PATHNAME_LEN_MAX];
	long current = 0;
	int res, amount, i, sum = 0, sum1 = 0, sum2=0;
	ListFiles_t list[JOIN_GRACE_TEMPSUBFILES_MAX];
	
	jg_prepairing(node->left, Descriptor, attr);//   .

	//  
	while (current < Descriptor->Sizelistfiles)
	{
		if (Descriptor->listfiles[current].statinfo.filesize > floor(((float)(Descriptor->table.size)/100)*70))
		{
			handle = fopen(Descriptor->listfiles[current].Filename, "rt");
			assert(handle != NULL);
			BYTECPY(filename, Descriptor->listfiles[current].Filename, strlen(Descriptor->listfiles[current].Filename)+1);
			//  
			amount = createrangefiles(Descriptor, list, current);
			//  
			distrib_tuples(list, handle, node->left->ary, attr, amount);
			//            1  
			while (current < Descriptor->Sizelistfiles - 1)	{
				BYTECPY(&Descriptor->listfiles[current], &Descriptor->listfiles[current+1], sizeof(Descriptor->listfiles[current+1]));
				current++;
			}

			Descriptor->Sizelistfiles = Descriptor->Sizelistfiles + amount - 1;
			assert(Descriptor->listfiles[current].statinfo.samplingsize < JOIN_GRACE_TEMPFILES_MAX);

			//     
			i = 0;
			while (current< Descriptor->Sizelistfiles) {
				BYTECPY(&Descriptor->listfiles[current], &list[i], sizeof(list[i]));
				current++;
				i++;
			}
			///////////////////////////////////////////
			res = fclose(handle);
			assert(res != EOF);
			res = remove(filename); //   .
			assert(res != EOF);
			current = 0;
		}
		else 
			current++;
	}
	//  .

	/*        (   ) */
	for (i = 0; i<Descriptor->Sizelistfiles; i++) {
		jg_CreateFilename(Descriptor->WhoAmI, Descriptor->listfiles[i].ProbeFilename);
		Descriptor->listfiles[i].Probehandle = fopen(Descriptor->listfiles[i].ProbeFilename, "wt");
		assert(Descriptor->listfiles[i].Probehandle != NULL);
		Descriptor->listfiles[i].statinfo.Probefilesize = 0;
	}

	br_Next(node->right);
	while (!node->right->isEOF) {

		i = getpos(Descriptor->listfiles, node->right->result[attr], Descriptor->Sizelistfiles);
		if (i != NIL) {
			res = jg_puttuple(Descriptor->listfiles[i].Probehandle, node->right->result, node->right->ary);
			assert(res != EOF);
			Descriptor->listfiles[i].statinfo.Probefilesize++;
		}
		br_Next(node->right);
	}
	
	for (i = 0; i<Descriptor->Sizelistfiles; i++) {
		res = fclose(Descriptor->listfiles[i].Probehandle);
		assert(res != EOF);
	}
	/* ------------------------------------------------------------------------ */
	listoptimizer(Descriptor->listfiles, &Descriptor->Sizelistfiles);	

	return;
} /* jg_ partioning() */

/*   .*/
static void ResetJOIN_Grace(treenode_t * node)
{
	assert(node != NULL);
	
	return;
}

/*   .      [0..size]*/
long hash_func1(long k, long size)
{
#define A ((sqrt(5)-1)/2)
	double temp = k*A;

	return (long)floor(size*(temp - floor(temp)));

#undef A
} /* hash_func1() */

/*   .       [0..size]*/
long hash_func2(long k, long size)
{
#define A ((sqrt(5)-1)/2)
	double temp = (k+5)*A;
	long x;

	x = (long)floor(size*(temp - floor(temp)));
	if (x%2 == 0)
		x = (x+1)%size;
	return x;

#undef A
} /* hash_func2() */

/*
 *  - table   list[current],  ary 
 *    attr.
 *    .        
 *    ,
 *        ,     
 *        
*/
void maketable(TupleBuffer_t *table, ListFiles_t *list, long current, int ary, int attr)
{
	int j, res;
	long h1, h2, hash;
	tuple_t tuple;
	buf_t *currentelem;

	for (j = 0; j < table->size; j++) {
		table->buffer[j].tuple[attr] = NIL;
		table->buffer[j].next = NULL;
	}

	list[current].handle = fopen(list[current].Filename, "rt");
	assert(list[current].handle != NULL);

	while (!feof(list[current].handle)) {
		res = jg_gettuple(list[current].handle, tuple, ary);

		h1 = hash_func1(tuple[attr], table->size);
		h2 = hash_func2(tuple[attr], table->size);
		currentelem = NULL;
		j = 0;
		hash = (h1+j*h2)%table->size;
		while (table->buffer[hash].tuple[attr] != NIL) {
			if ((currentelem == NULL) && (table->buffer[hash].tuple[attr] == tuple[attr])) {
				currentelem = &table->buffer[hash];
				while (currentelem->next != NULL) {
					currentelem = currentelem->next;
				}
			}
			j++;
			hash = (h1+j*h2)%table->size;
		}
		if (currentelem !=NULL)
			currentelem->next = &table->buffer[hash];

		TUPLECPY(table->buffer[hash].tuple, tuple, ary);
		res++;

		assert(res != table->size);
	}
	
	res = fclose(list[current].handle);
	assert(res != EOF);
	res = remove(list[current].Filename);
	assert (res != -1);

	return;
}

/*
 *    .
*/
static tupleptr_t NextJOIN_Grace(treenode_t * node)
{
	JOIN_GRACE_cargo_t *Descriptor;
	int attr, res, i, j;
	tuple_t tuple;
	long h1, h2, hash;

	assert(node != NULL);

	Descriptor = (JOIN_GRACE_cargo_t *) JOIN_CARGO(node, supercargo);
	assert(Descriptor != NULL);
	attr = JOIN_CARGO(node, attribute);
	//       .
	if (Descriptor->ResetSign) {
		Descriptor->ResetSign = FALSE;
		if (WHOAMI == 0)
			printf("\n--------------GRACE JOIN--------------\n");
		srand(1);
		Descriptor->table.size = getsizetable(JOIN_GRACE_BUFLEN_MAX);
		jg_partitioning(node, Descriptor, attr);
		Descriptor->listfiles[0].Probehandle = fopen(Descriptor->listfiles[0].ProbeFilename, "rt");
		assert(Descriptor->listfiles[0].Probehandle != NULL);
		Descriptor->currentfile = 0;
		maketable(&Descriptor->table, Descriptor->listfiles, Descriptor->currentfile, node->left->ary, attr);
		Descriptor->table.current = NULL;
	}
	/////////////////
	//      ,     ,    
	//   -
	tuple[attr] = NIL;
	
	if (Descriptor->table.current == NULL) {
	//       ...   
	while (tuple[attr] == NIL) {

		if (feof(Descriptor->listfiles[Descriptor->currentfile].Probehandle))
		//    ,   ...
		{
			res = fclose(Descriptor->listfiles[Descriptor->currentfile].Probehandle);
			assert(res != EOF);
			res = remove(Descriptor->listfiles[Descriptor->currentfile].ProbeFilename);
			assert (res != -1);

			if (Descriptor->currentfile == Descriptor->Sizelistfiles-1) { 
			//       ...
				RESCPY(node, EOFtuple);

				return node->result;
			}

			Descriptor->currentfile++;
			maketable(&Descriptor->table, Descriptor->listfiles, Descriptor->currentfile, node->left->ary, attr);
			Descriptor->listfiles[Descriptor->currentfile].Probehandle = fopen(Descriptor->listfiles[Descriptor->currentfile].ProbeFilename, "rt");
			assert(Descriptor->listfiles[Descriptor->currentfile].Probehandle);
		}

		jg_gettuple(Descriptor->listfiles[Descriptor->currentfile].Probehandle, Descriptor->Probetuple, node->right->ary);
		//   
		h1 = hash_func1(Descriptor->Probetuple[attr], Descriptor->table.size); 
		h2 = hash_func2(Descriptor->Probetuple[attr], Descriptor->table.size); 
		j = 0;
		hash = (h1 + j * h2)%Descriptor->table.size;
		while (Descriptor->table.buffer[hash].tuple[attr] != NIL) {
			if (Descriptor->table.buffer[hash].tuple[attr] ==  Descriptor->Probetuple[attr]) {
				Descriptor->table.current = Descriptor->table.buffer[hash].next;
				TUPLECPY(tuple, Descriptor->table.buffer[hash].tuple, node->right->ary);

				break;
			}
			j++;
			hash = (h1 + j * h2)%Descriptor->table.size;
		}
		//  .   ,     
	} //while
	} // if
	else {
		TUPLECPY(tuple, Descriptor->table.current->tuple, node->right->ary);
		Descriptor->table.current = Descriptor->table.current->next;
	}

	assert(Descriptor->Probetuple[attr] == tuple[attr]);
	//           . 
	//  ,       .
	if (Descriptor->listfiles[Descriptor->currentfile].ReverseSign) {
	//    
	for (i=0; i<node->left->ary; i++)
		node->result[i]=tuple[i];
	for (i=0; i<JOIN_CARGO(node, attribute); i++)
		node->result[node->left->ary+i]=Descriptor->Probetuple[i];
	for (i=JOIN_CARGO(node, attribute)+1; i<node->right->ary; i++)
		node->result[node->left->ary+i-1]=Descriptor->Probetuple[i];
	}
	else {
		for (i=0; i<node->right->ary; i++)
			node->result[i]=Descriptor->Probetuple[i];
		for (i=0; i<JOIN_CARGO(node, attribute); i++)
			node->result[node->left->ary+i]=tuple[i];
		for (i=JOIN_CARGO(node, attribute)+1; i<node->right->ary; i++)
			node->result[node->right->ary+i-1]=tuple[i];
	}
	return node->result;
}
#undef RESCPY
#undef TUPLECPY
#undef BYTECPY
/* END */