/*
  JOIN ( )  , 
       .
.. 
@version 1.0
*/

#include "bracket.h"	//  
#include "dbparams.h"	//  
#include "jhash.h"	    // 
#include "dbmgr.h"
#include <string.h>
#include <stdlib.h> 
#include <stdio.h>
#include "assert.h"
#include "math.h"

/* INTERFACE */

/* Types: */

/*
 *  -.
 * tuple - .
 * next -     ,      .
*/
struct _buf{
	tuple_t tuple;
	struct _buf *next;
};

typedef struct _buf buf_t;

/*
 * ,  -.
 * size -  -.
 * buffer -   -.
 * current -      ,         .
 * 
*/
typedef struct {
	buf_t buffer[JOIN_HASH_BUFLEN_MAX];	
	long size;
	buf_t *current;
} TupleBuffer_t;

/*
 *   .
 * table -   ,  -.
 * isleft - ,   . TRUE - , FALSE - .
*/
typedef struct {
	TupleBuffer_t table;
	int isleft;
	treenode_t *build, *probe;
} JOIN_HASH_cargo_t;

/* Function prototypes: */

/**
       
   JOIN_HASH
(  ,       
   ). 
@memo   JOIN_HASH
@param node   
*/
void AssignJOIN_Hash(treenode_t * node, int isleft);

/*
@memo    JOIN  
@param node     
*/
static void ResetJOIN_Hash(treenode_t * node);

/*
  JOIN  
@param node     
@return        , 
  ;     --   -EOF.
*/
static tupleptr_t NextJOIN_Hash(treenode_t * node);

/*
 *  - table   node.
*/
void jh_maketable(treenode_t *node, TupleBuffer_t *table, int attr);

long jh_hash_func1(long k, long size);

long jh_hash_func2(long k, long size);

long gettablesize(long size);
/* Definitions: */

#define JOIN_HASH_CARGO(treenode, attr)	((JOIN_HASH_cargo_t *) JOIN_CARGO(treenode, supercargo))->attr
/*
 *  .
*/
#define TUPLECPY(dest, src, ary) memcpy(dest, src, ary*sizeof(domain_t))
#define RESCPY(node, tuple)	memcpy(node->result, tuple, node->ary*sizeof(domain_t))

/* IMPLEMENTATION */
/*
 *    .
*/
void AssignJOIN_Hash(treenode_t * node, int isleft)
{
	assert(node!=NULL);
	assert(node->rel_operation==JOIN);
	
	node->phys_operation=JOIN_HASH;
	node->nextfn=(next_fn_t) NextJOIN_Hash;
	node->resetfn=(reset_fn_t) ResetJOIN_Hash;
	JOIN_CARGO(node, supercargo)=(JOIN_HASH_cargo_t *) calloc(1, sizeof(JOIN_HASH_cargo_t));
	assert(JOIN_CARGO(node, supercargo)!=NULL);
	if (isleft)
		JOIN_HASH_CARGO(node, isleft) = TRUE;
	else
		JOIN_HASH_CARGO(node, isleft) = FALSE;

	return;
}

static void ResetJOIN_Hash(treenode_t * node)
{
	assert(node != NULL);

	return;
}

static tupleptr_t NextJOIN_Hash(treenode_t * node)
{
	JOIN_HASH_cargo_t *Descriptor;
	int attr, i, j;
	long h1, h2, hash;
	tuple_t tuple;
	
	assert(node != NULL);
	Descriptor = (JOIN_HASH_cargo_t *) JOIN_CARGO(node, supercargo);
	assert(Descriptor != NULL);
	attr = JOIN_CARGO(node, attribute);
	if (node->left->isBOF) {
	if (JOIN_HASH_CARGO(node, isleft)) {
		Descriptor->build = node->left;
		Descriptor->probe = node->right;
	}
	else {
		Descriptor->build = node->right;
		Descriptor->probe = node->left;
	}
		Descriptor->table.size = gettablesize(JOIN_HASH_BUFLEN_MAX);
		Descriptor->table.current = NULL;
		jh_maketable(Descriptor->build, &Descriptor->table, attr);
	}

	tuple[attr] = NIL;

	if (Descriptor->table.current == NULL) {
	while (tuple[attr] == NIL) {
		br_Next(Descriptor->probe);
		if (Descriptor->probe->isEOF) {
			RESCPY(node, EOFtuple);

			return node->result;
		}
		else {
			h1 = jh_hash_func1(Descriptor->probe->result[attr], Descriptor->table.size);
			h2 = jh_hash_func2(Descriptor->probe->result[attr], Descriptor->table.size);
			j = 0;
			hash = (h1 + j * h2)%Descriptor->table.size;
			while (Descriptor->table.buffer[hash].tuple[attr] != NIL) {
				if (Descriptor->table.buffer[hash].tuple[attr] ==  Descriptor->probe->result[attr]) {
					TUPLECPY(tuple, Descriptor->table.buffer[hash].tuple, Descriptor->probe->ary);
					Descriptor->table.current = Descriptor->table.buffer[hash].next;

					break;
				}
				j++;
				hash = (h1 + j * h2)%Descriptor->table.size;
			} // while
		} //else
	}
	}
	else {
		TUPLECPY(tuple, Descriptor->table.current->tuple, Descriptor->probe->ary);
		Descriptor->table.current = Descriptor->table.current->next;
	}
	if (JOIN_HASH_CARGO(node, isleft)) {
	//    
	for (i=0; i<Descriptor->build->ary; i++)
		node->result[i]=tuple[i];
	for (i=0; i<JOIN_CARGO(node, attribute); i++)
		node->result[Descriptor->build->ary+i]=Descriptor->probe->result[i];
	for (i=JOIN_CARGO(node, attribute)+1; i<Descriptor->probe->ary; i++)
		node->result[Descriptor->build->ary+i-1]=Descriptor->probe->result[i];
	}
	else {
	//    
	for (i=0; i<Descriptor->probe->ary; i++)
		node->result[i]=Descriptor->probe->result[i];
	for (i=0; i<JOIN_CARGO(node, attribute); i++)
		node->result[Descriptor->probe->ary+i]=tuple[i];
	for (i=JOIN_CARGO(node, attribute)+1; i<Descriptor->build->ary; i++)
		node->result[Descriptor->probe->ary+i-1]=tuple[i];
	}
	return node->result;
}

/*
 *   -:
 * 1)         (NIL);    .
 * 2)  ,  -.
 * 3)             ,    , 
                  
 *      ,       .
 * 4)  .2   EOF   node.
*/
void jh_maketable(treenode_t *node, TupleBuffer_t *table, int attr)
{
	int j, res;
	long h1, h2, hash;
	buf_t *currentelem = NULL;

	for (j = 0; j < table->size; j++) {
		table->buffer[j].tuple[attr] = NIL;
		table->buffer[j].next = NULL;
	}

	res = 0;
	br_Next(node);
	while (!node->isEOF) {
	
	
		h1 = jh_hash_func1(node->result[attr], table->size);
		h2 = jh_hash_func2(node->result[attr], table->size);
		currentelem = NULL;
		j = 0;
		hash = (h1+j*h2)%table->size;
		while (table->buffer[hash].tuple[attr] != NIL) {
			if ((currentelem == NULL) && (table->buffer[hash].tuple[attr] == node->result[attr])) {
				currentelem = &table->buffer[hash];
				while (currentelem->next != NULL) {
					currentelem = currentelem->next;
				}
			}
			j++;
			hash = (h1+j*h2)%table->size;
		}
		if (currentelem !=NULL)
			currentelem->next = &table->buffer[hash];
		table->buffer[hash].next = NULL;
		TUPLECPY(table->buffer[hash].tuple, node->result, node->ary);
		res++;

		assert(res != table->size);
		br_Next(node);
	}	

	return;
}

/*
 *   -.
 *  - -  2^n,    .
*/
long gettablesize(long size)
{
	long m;
	
	m = 1;
	while (m <= size)
		m = m<<1;
	m = m>>1;

	return m;
}

/*
 *  -.   (0...2^n).   .
*/
long jh_hash_func1(long k, long size)
{
#define A ((sqrt(5)-1)/2)
	double temp = k*A;

	return (long)floor(size*(temp - floor(temp)));

#undef A
} /* hash_func1() */

/*
 *  -.      (0...2^n).   .
*/
long jh_hash_func2(long k, long size)
{
#define A ((sqrt(5)-1)/2)
	double temp = (k+5)*A;
	long x;

	x = (long)floor(size*(temp - floor(temp)));
	if (x%2 == 0)
		x = (x+1)%size;
	return x;

#undef A
} /* hash_func2() */
#undef RESCPY
#undef TUPLECPY
/* END */